#include "win32l.h"
#undef NOUSER
#include <windows.h>
#include "kmp_com.h"
#include "xstdlib.h"
#include "fmpmdif.h"

#define KPI_VERSION 6
#define KPI_VERSIONS "r6"

struct KMP_CTX_TAG
{
	HMODULE hDll;
	IFMPMD *pifmpmd;
	int ifver;
	int isfmp;
	int rptr;
	int srate;
	int loopcount;
	int length;
	int looplength;
	union {
		struct {
			int fmvoldown;
			int ssgvoldown;
			int rhythmvoldown;
			int adpcmvoldown;
			int ppzvoldown;
			int ppsuse;
			int rhythmwithssgeffect;
			int pmd86pcmmode;
		} pmd;
		struct {
			int fmvoldown;
			int ssgvoldown;
			int rhythmvoldown;
			int adpcmvoldown;
			int ppzvoldown;
			int adpcmppz8emulate;
		} fmp;
	} options;
};

#define RLIMIT 4
/*
 ɕs邽߁A
 炩ߕIuWFNgmۂĂ܂B
  1:tp
  2:NXtF[hp
  3:tԎ擾p
  4:\
 ȂȂꍇ͉t܂AʂɊmۂ܂B
*/

static struct {
	int mode;
	int srate_default;
	HMODULE hDll[2];
	int alloc[2];
	int inuse[2];
	IFMPMD *pif[2][RLIMIT];
	int srate[2][RLIMIT];
} rwork = {
	1,
	44100,
	{0, 0},
	{0, 0},
	{0, 0},
	{{0, 0, 0, 0}, {0, 0, 0, 0}},
};

static char iniFilePath[MAX_PATH] = "";
static const char *sectionname = "FMPMD";
static const char *valuename_sr = "SampleRate";
static const char *valuename_l = "Loop";
static const char *valuename_erm = "EnableReentrant";
static const char *valuename_emo = "EnableMemoOptions";
static const char *valuename_dfo = "DefaultFMPOptions";
static const char *valuename_dpo = "DefaultPMDOptions";
static const char *valuename_rp = "RhythmPath";
static const char *valuename_o = "Options";

static const char *regsubkey1 = "Software\\Kobarin's Soft\\KbMedia Player\\Path";
static const char *regsubkey2 = "Software\\Kobarin's Soft\\KbMIDI Player\\SearchPath";

static UINT GetPrivateProfileIntEx(LPCTSTR lpAppName, LPCTSTR lpKeyName, INT nDefault, LPCTSTR lpFileName)
{
	UINT ret;
	TCHAR szBuffer[32];
	ret = GetPrivateProfileInt(lpAppName, lpKeyName, nDefault, lpFileName);
	wsprintf(szBuffer, "%d", ret);
	WritePrivateProfileString(lpAppName, lpKeyName, szBuffer, lpFileName);
	return ret;
}

static BOOL GetPrivateProfileBoolEx(LPCTSTR lpAppName, LPCTSTR lpKeyName, BOOL bDefault, LPCTSTR lpFileName)
{
	BOOL ret;
	TCHAR szBuffer[32];
	if (GetPrivateProfileString(lpAppName, lpKeyName, bDefault ? "True" : "False", szBuffer, 32, lpFileName))
	{
		/* 擪mF */
		ret = szBuffer[0] == 'T' || szBuffer[0] == 't';
	}
	else
		ret = bDefault;
	WritePrivateProfileString(lpAppName, lpKeyName, ret ? "True" : "False", lpFileName);
	return ret;
}

static const char *dllname[] = { PMD_DLL_NAME, FMP_DLL_NAME };
static const char *giprocname[] = { PMD_GETINTERFACEVERSION_PROC_NAME, FMP_GETINTERFACEVERSION_PROC_NAME };
static const char *gvprocname[] = { PMD_GETVERSION_PROC_NAME, FMP_GETVERSION_PROC_NAME };
static const char *ciprocname[] = { PMD_COCREATEINSTANCE_PROC_NAME, FMP_COCREATEINSTANCE_PROC_NAME };
static const FCCLSID *clsids[] = { &FCCLSID_PMDWIN, &FCCLSID_WINFMP };
static const unsigned minver[] = { PMD_INTERFACEVERSION, FMP_INTERFACEVERSION };
static const unsigned maxver[] = { PMD_INTERFACEVERSION_LIMIT, FMP_INTERFACEVERSION_LIMIT };

static HMODULE LoadDll(int isfmp, int *pifver)
{
	HMODULE ret;
	char path[MAX_PATH];
	unsigned pos;
	pos = XDIVPATH(path, XGETMODULEPATH(), MAX_PATH);
	XSTRCPYN(path + pos, dllname[isfmp], MAX_PATH - pos);
	ret = LoadLibrary(path);
	if (!ret)
	{
		XSTRCPYN(path, dllname[isfmp], MAX_PATH);
		ret = LoadLibrary(path);
	}
	if (ret)
	{
		GETINTERFACEVERSION_PROC_TYPE *pfngiv;
		pfngiv = (GETINTERFACEVERSION_PROC_TYPE *)GetProcAddress(ret, giprocname[isfmp]);
		if (pfngiv)
		{
			unsigned iv;
			iv = pfngiv();
			if (pifver) *pifver = iv;
			if (minver[isfmp] <= iv && iv <= maxver[isfmp]) return ret;
		}
		FreeLibrary(ret);
	}
	return 0;
}

static unsigned GetDllVersion(int isfmp)
{
	unsigned ret;
	HMODULE hDll;
	GETVERSION_PROC_TYPE *pfngv;
	hDll = LoadDll(isfmp, 0);
	if (!hDll) return 0;
	pfngv = (GETVERSION_PROC_TYPE *)GetProcAddress(hDll, gvprocname[isfmp]);
	if (!pfngv) return 0;
	ret = pfngv();
	FreeLibrary(hDll);
	return ret;
}

static char *RegGetString(HKEY hkeyroot, const char *subkey, const char *valuename)
{
	HKEY hkey;
	char *buf = 0;
	if (RegOpenKeyEx(hkeyroot, subkey, 0, KEY_QUERY_VALUE, &hkey) == ERROR_SUCCESS)
	{
		DWORD type, size;
		if (RegQueryValueEx(hkey, valuename, 0, &type, 0, &size) == ERROR_SUCCESS && size > 1 && type == REG_SZ)
		{
			buf = XMALLOC(size);
			if (buf)
			{
				if (RegQueryValueEx(hkey, valuename, 0, &type, buf, &size) != ERROR_SUCCESS || size <= 1 || type != REG_SZ)
				{
					XFREE(buf);
					buf = 0;
				}
			}
		}
		RegCloseKey(hkey);
	}
	return buf;
}

static void InitRhythmPath(char *rhythmpath, unsigned len)
{
	const char *regvaluename = "2608.WAV";
	char *rp = 0;
	rp = RegGetString(HKEY_CURRENT_USER, regsubkey1, regvaluename);
	if (!rp) rp = RegGetString(HKEY_CURRENT_USER, regsubkey2, regvaluename);
	if (rp)
	{
		XSTRCPYN(rhythmpath, rp, len);
	}
	else if (GetPrivateProfileString(sectionname, valuename_rp, "", rhythmpath, sizeof(rhythmpath) - 1, iniFilePath) == 0)
	{
		XDIVPATH(rhythmpath, XGETMODULEPATH(), len);
	}
	if (rp) XFREE(rp);
}

static void InitSub(int isfmp)
{
	XGLOBALLOCK();
	if (rwork.mode && rwork.alloc[isfmp] == 0)
	{
		if (!rwork.hDll[isfmp]) rwork.hDll[isfmp] = LoadDll(isfmp, 0);
		if (rwork.hDll[isfmp])
		{
			COCREATEINSTANCE_PROC_TYPE *pfnci;
			pfnci = (COCREATEINSTANCE_PROC_TYPE *)GetProcAddress(rwork.hDll[isfmp], ciprocname[isfmp]);
			if (pfnci)
			{
				int i;
				char rhythmpath[MAX_PATH];
				InitRhythmPath(rhythmpath, sizeof(rhythmpath));
				for (i = 0; i < RLIMIT; i++)
				{
					FCHRESULT hr;
					hr = pfnci(clsids[isfmp], 0, FCCLSCTX_ALL, &FCIID_IFMPMD, (void **)&rwork.pif[isfmp][i]);
					if (hr >= 0)
					{
						rwork.pif[isfmp][i]->lpVtbl->init(rwork.pif[isfmp][i], rhythmpath);
						rwork.pif[isfmp][i]->lpVtbl->setpcmrate(rwork.pif[isfmp][i], rwork.srate_default);
						rwork.pif[isfmp][i]->lpVtbl->setfmcalc55k(rwork.pif[isfmp][i], 1);
						rwork.pif[isfmp][i]->lpVtbl->setppzinterpolation(rwork.pif[isfmp][i], 1);
						rwork.srate[isfmp][i] = rwork.srate_default;
						rwork.alloc[isfmp] |= 1 << i;
					}
				}
			}
		}
	}
	XGLOBALUNLOCK();
}

static void DeinitSub(int isfmp)
{
	XGLOBALLOCK();
	if (rwork.inuse[isfmp] == 0)
	{
		int i;
		for (i = 0; i < RLIMIT; i++)
		{
			if (rwork.alloc[isfmp] & (1 << i))
			{
				rwork.alloc[isfmp] &= ~(1 << i);
				rwork.pif[isfmp][i]->lpVtbl->Release(rwork.pif[isfmp][i]);
			}
		}
		if (rwork.hDll[isfmp])
		{
			FreeLibrary(rwork.hDll[isfmp]);
			rwork.hDll[isfmp] = 0;
		}
	}
	XGLOBALUNLOCK();
}

void Init(void)
{
	/* Kv͂܂ */
}

void Deinit(void)
{
	/* Kv͂܂ */
}

static FCBOOL CreateInstance(KMP_CTX *pctx, char *rhythmpath)
{
	pctx->pifmpmd = 0;
	pctx->hDll = LoadDll(pctx->isfmp, &pctx->ifver);
	if (pctx->hDll)
	{
		COCREATEINSTANCE_PROC_TYPE *pfnci;
		pfnci = (COCREATEINSTANCE_PROC_TYPE *)GetProcAddress(pctx->hDll, ciprocname[pctx->isfmp]);
		if (pfnci)
		{
			if (rwork.mode)
			{
				int i;
				InitSub(pctx->isfmp);
				XGLOBALLOCK();
				for (i = 0; i < RLIMIT; i++)
				{
					if ((rwork.alloc[pctx->isfmp] ^ rwork.inuse[pctx->isfmp]) & (1 << i))
					{
						pctx->rptr = 1 << i;
						rwork.inuse[pctx->isfmp] |= 1 << i;
						pctx->pifmpmd = rwork.pif[pctx->isfmp][i];
						pctx->srate = rwork.srate[pctx->isfmp][i];
						break;
					}
				}
				XGLOBALUNLOCK();
			}
			if (!pctx->pifmpmd)
			{
				FCHRESULT hr;
				hr = pfnci(clsids[pctx->isfmp], 0, FCCLSCTX_ALL, &FCIID_IFMPMD, (void **)&pctx->pifmpmd);
				if (hr >= 0)
				{
					pctx->pifmpmd->lpVtbl->init(pctx->pifmpmd, rhythmpath);
					pctx->pifmpmd->lpVtbl->setpcmrate(pctx->pifmpmd, pctx->srate);
					pctx->pifmpmd->lpVtbl->setfmcalc55k(pctx->pifmpmd, 1);
					pctx->pifmpmd->lpVtbl->setppzinterpolation(pctx->pifmpmd, 1);
				}
				else pctx->pifmpmd = 0;
			}
		}
	}
	return pctx->pifmpmd != 0;
}

static void ReleaseInstance(KMP_CTX *pctx)
{
	if (pctx->rptr)
	{
		XGLOBALLOCK();
		pctx->pifmpmd = 0;
		rwork.inuse[pctx->isfmp] &= ~pctx->rptr;
		XGLOBALUNLOCK();
		/* DeinitSub(pctx->isfmp); */
	}
	else
	{
		if (pctx->pifmpmd)
		{
			pctx->pifmpmd->lpVtbl->Release(pctx->pifmpmd);
			pctx->pifmpmd = 0;
		}
	}
	if (pctx->hDll)
	{
		FreeLibrary(pctx->hDll);
		pctx->hDll = 0;
	}
}

static int abortcheck(KMP_CTX *pctx, int errcode)
{
	switch (errcode)
	{
	case FMPMD_OK:
	case FMPMD_WARNING_PPC_ALREADY_LOAD:
	case FMPMD_WARNING_P86_ALREADY_LOAD:
	case FMPMD_WARNING_PPS_ALREADY_LOAD:
	case FMPMD_WARNING_PPZ1_ALREADY_LOAD:
	case FMPMD_WARNING_PPZ2_ALREADY_LOAD:
		return 0;
	}
	return 1;
}

char *XSTRTOK(char *p, const char *s)
{
	while (*p)
	{
		int i;
		for (i = 0; s[i]; i++)
		{
			if (*p == s[i])
			{
				*p = 0;
				return p + 1;
			}
		}
		p += XMBLEN(p);
	}
	return p;
}

#define	MAX_PCMDIR				 64
static void SetPCMDirectories(KMP_CTX *pctx)
{
	static const char *regvaluename = "PVI";
	static const char *sepp = ";";
	char *dirs[MAX_PCMDIR + 1];
	char *buf1 = 0;
	int i = 0;
	buf1 = RegGetString(HKEY_CURRENT_USER, regsubkey1, regvaluename);
	if (!buf1) buf1 = RegGetString(HKEY_CURRENT_USER, regsubkey2, regvaluename);
	if (buf1)
	{
		char *p = buf1;
		while (*p)
		{
			if (i < MAX_PCMDIR) dirs[i++] = p;
			p = XSTRTOK(p, sepp);
		}
	}
	dirs[i] = 0;
	if (i) pctx->pifmpmd->lpVtbl->setpcmdir(pctx->pifmpmd, dirs);
	if (buf1) XFREE(buf1);
}

static char *strnum(char *p, int *ret)
{
	int ok = 0, num = 0;
	while (1)
	{
		switch (*p)
		{
		case '0': case '1': case '2': case '3': case '4':
		case '5': case '6': case '7': case '8': case '9':
			ok = 1;
			num = num * 10 + *p++ - '0';
			continue;
		default:
			if (ok) *ret = num;
			return p;
		}
	}
}

static void GetPMDOptions(KMP_CTX *pctx, char *p, int mode)
{
	while (p[0])
	{
		static const char *sepp = " \t\n\r\"";
		if (mode || *p == '/')
		{
			if (!mode) p++;
			switch (p[0])
			{
			case 'D':
				p++;
				switch (p[0])
				{
				case 'F':
					p = strnum(p + 1, &pctx->options.pmd.fmvoldown);
					continue;
				case 'S':
					p = strnum(p + 1, &pctx->options.pmd.ssgvoldown);
					continue;
				case 'P':
					p = strnum(p + 1, &pctx->options.pmd.adpcmvoldown);
					continue;
				case 'R':
					p = strnum(p + 1, &pctx->options.pmd.rhythmvoldown);
					continue;
				case 'Z':
					p = strnum(p + 1, &pctx->options.pmd.ppzvoldown);
					continue;
				}
				break;
			case 'N':
				p++;
				if (p[0] == '-')
				{
					p++;
					pctx->options.pmd.rhythmwithssgeffect = 1;
				}
				else
					pctx->options.pmd.rhythmwithssgeffect = 0;
				continue;
			case 'P':
				p++;
				if (p[0] == '-')
				{
					p++;
					pctx->options.pmd.ppsuse = 1;
				}
				else
					pctx->options.pmd.ppsuse = 0;
				continue;
			case 'S':
				p++;
				if (p[0] == '-')
				{
					p++;
					pctx->options.pmd.pmd86pcmmode = 0;
				}
				else
					pctx->options.pmd.pmd86pcmmode = 1;
				continue;
			}
		}
		p = XSTRTOK(p, sepp);
	}
}

static void GetFMPOptions(KMP_CTX *pctx, char *p, int mode)
{
	while (p[0])
	{
		static const char *sepp = " \t\n\r\"";
		if (mode || *p == '/' || *p == '-')
		{
			if (!mode) p++;
			switch (p[0])
			{
			case '/':
				p++;
				if (!mode)
				{
					if (p[0] == '-')
					{
						p++;
						pctx->options.fmp.adpcmppz8emulate = 0;
					}
					else
						pctx->options.fmp.adpcmppz8emulate = 1;
				}
				continue;
			case 'V':
				p++;
				switch (p[0])
				{
				case 'F':
					p = strnum(p + 1, &pctx->options.fmp.fmvoldown);
					continue;
				case 'S':
					p = strnum(p + 1, &pctx->options.fmp.ssgvoldown);
					continue;
				case 'P':
					p = strnum(p + 1, &pctx->options.fmp.adpcmvoldown);
					continue;
				case 'R':
					p = strnum(p + 1, &pctx->options.fmp.rhythmvoldown);
					continue;
				case 'Z':
					p = strnum(p + 1, &pctx->options.fmp.ppzvoldown);
					continue;
				}
				break;
			}
		}
		p = XSTRTOK(p, sepp);
	}
}

static void GetOptions(KMP_CTX *pctx, const char *path, const char *fname)
{
	char options[128];
	int emo;

	/* B@\:memoɃIvVLq */
	emo = GetPrivateProfileInt(sectionname, valuename_emo, 0, iniFilePath);

	switch (pctx->isfmp)
	{
	case 0:	/* PMD */
		/* default options */
		pctx->options.pmd.fmvoldown = 16;
		pctx->options.pmd.ssgvoldown = 0;
		pctx->options.pmd.rhythmvoldown = 0;
		pctx->options.pmd.adpcmvoldown = 0;
		pctx->options.pmd.ppzvoldown = 0;
		pctx->options.pmd.ppsuse = 1;
		pctx->options.pmd.rhythmwithssgeffect = 0;
		pctx->options.pmd.pmd86pcmmode = 0;
		if (GetPrivateProfileString(fname, valuename_o, "", options, sizeof(options) - 1, iniFilePath) != 0 || GetPrivateProfileString(sectionname, valuename_dpo, "", options, sizeof(options) - 1, iniFilePath) != 0)
			GetPMDOptions(pctx, options, 0);
		if (emo)
		{
			IPMDWIN *ipmdwin;
			if (pctx->pifmpmd->lpVtbl->QueryInterface(pctx->pifmpmd, &FCIID_IPMDWIN, (void **)&ipmdwin) >= 0)
			{
				int i;
				for (i = 4; i < 4 + 3; i++)
				{
					if (ipmdwin->lpVtbl->getmemo3(ipmdwin, options, 0, 0, i) == FMPMD_OK)
						GetPMDOptions(pctx, options, 1);
				}
				ipmdwin->lpVtbl->Release(ipmdwin);
			}
		}
		break;
	case 1:	/* FMP */
		pctx->options.fmp.fmvoldown = 0;
		pctx->options.fmp.ssgvoldown = 0;
		pctx->options.fmp.rhythmvoldown = 0;
		pctx->options.fmp.adpcmvoldown = 0;
		pctx->options.fmp.ppzvoldown = 0;
		pctx->options.fmp.adpcmppz8emulate = 0;
		if (GetPrivateProfileString(fname, valuename_o, "", options, sizeof(options) - 1, iniFilePath) != 0 || GetPrivateProfileString(sectionname, valuename_dfo, "", options, sizeof(options) - 1, iniFilePath) != 0)
			GetFMPOptions(pctx, options, 0);
		if (emo)
		{
			IWINFMP *iwinfmp;
			if (pctx->pifmpmd->lpVtbl->QueryInterface(pctx->pifmpmd, &FCIID_IWINFMP, (void **)&iwinfmp) >= 0)
			{
				char memobuf[3][80+2];
				char *memoarray[3];
				memoarray[0] = memobuf[0];
				memoarray[1] = memobuf[1];
				memoarray[2] = memobuf[2];
				if (iwinfmp->lpVtbl->getcomment3(iwinfmp, memoarray ,0, 0) == FMPMD_OK)
				{
					GetFMPOptions(pctx, memobuf[0], 1);
					GetFMPOptions(pctx, memobuf[1], 1);
					GetFMPOptions(pctx, memobuf[2], 1);
				}
				iwinfmp->lpVtbl->Release(iwinfmp);
			}
		}
		break;
	}
}

void Close(KMP_CTX *pctx)
{
	ReleaseInstance(pctx);
	XFREE(pctx);
}

KMP_CTX *Open(const char *path, unsigned size, unsigned srate)
{
	KMP_CTX *pctx;
	pctx = XMALLOC(sizeof(KMP_CTX));
	while (pctx)
	{
		char rhythmpath[MAX_PATH];
		const char *fname;

		pctx->hDll = 0;
		pctx->pifmpmd = 0;
		pctx->rptr = 0;

		if (size)
		{
#if 0
			static const char *unknown = "unknown";
			static const char *checkid = "FMC";
			unsigned ofs;
			fname = unknown;
			ofs = path[0] + (path[1] << 8);
			if (ofs + 3 < size && !XMEMCMP((void *)(path + ofs), (void *)checkid, 3))
				pctx->isfmp = 1;
			else
				pctx->isfmp = 0;
#endif
			break;
		}
		else
		{
			const char *ext;
			fname = path + XDIVPATH(0, path, ~0);
			ext = path + XDELEXT(0, path, ~0);
			pctx->isfmp = (ext[0] == '.' && (ext[1] == 'O' || ext[1] == 'o'));
		}

		InitRhythmPath(rhythmpath, sizeof(rhythmpath));

		rwork.srate_default = pctx->srate = GetPrivateProfileIntEx(sectionname, valuename_sr, 44100, iniFilePath);
		if (srate) rwork.srate_default = pctx->srate = srate;
		pctx->loopcount = GetPrivateProfileInt(fname, valuename_l, GetPrivateProfileIntEx(sectionname, valuename_l, 2, iniFilePath), iniFilePath);

		if (!CreateInstance(pctx, rhythmpath)) break;

		SetPCMDirectories(pctx);

		if (abortcheck(pctx, pctx->pifmpmd->lpVtbl->music_load(pctx->pifmpmd, (char *)path))) break;

#if 0
		if (size)
			pctx->pifmpmd->lpVtbl->getlength(pctx->pifmpmd, (char *)0, &pctx->length, &pctx->looplength);
		else
#endif
		{
			pctx->pifmpmd->lpVtbl->getlength(pctx->pifmpmd, (char *)path, &pctx->length, &pctx->looplength);
			GetOptions(pctx, path, fname);
		}

		switch (pctx->isfmp)
		{
		case 0: /* PMD */
			{
				IPMDWIN *ipmdwin;
				if (pctx->pifmpmd->lpVtbl->QueryInterface(pctx->pifmpmd, &FCIID_IPMDWIN, (void **)&ipmdwin) >= 0)
				{
					ipmdwin->lpVtbl->setp86interpolation(ipmdwin, 1);
					ipmdwin->lpVtbl->setppsinterpolation(ipmdwin, 1);
					ipmdwin->lpVtbl->setfmvoldown(ipmdwin, pctx->options.pmd.fmvoldown);
					ipmdwin->lpVtbl->setssgvoldown(ipmdwin, pctx->options.pmd.ssgvoldown);
					ipmdwin->lpVtbl->setrhythmvoldown(ipmdwin, pctx->options.pmd.rhythmvoldown);
					ipmdwin->lpVtbl->setadpcmvoldown(ipmdwin, pctx->options.pmd.adpcmvoldown);
					ipmdwin->lpVtbl->setppzvoldown(ipmdwin, pctx->options.pmd.ppzvoldown);
					ipmdwin->lpVtbl->setppsuse(ipmdwin, pctx->options.pmd.ppsuse);
					ipmdwin->lpVtbl->setrhythmwithssgeffect(ipmdwin, pctx->options.pmd.rhythmwithssgeffect);
					ipmdwin->lpVtbl->setpmd86pcmmode(ipmdwin, pctx->options.pmd.pmd86pcmmode);
					ipmdwin->lpVtbl->Release(ipmdwin);
				}
			}
			break;
		case 1: /* FMP */
			{
				IWINFMP *iwinfmp;
				if (pctx->pifmpmd->lpVtbl->QueryInterface(pctx->pifmpmd, &FCIID_IWINFMP, (void **)&iwinfmp) >= 0)
				{
					iwinfmp->lpVtbl->setfmvoldown(iwinfmp, pctx->options.fmp.fmvoldown);
					iwinfmp->lpVtbl->setssgvoldown(iwinfmp, pctx->options.fmp.ssgvoldown);
					iwinfmp->lpVtbl->setrhythmvoldown(iwinfmp, pctx->options.fmp.rhythmvoldown);
					iwinfmp->lpVtbl->setadpcmvoldown(iwinfmp, pctx->options.fmp.adpcmvoldown);
					iwinfmp->lpVtbl->setppzvoldown(iwinfmp, pctx->options.fmp.ppzvoldown);
					if (pctx->ifver >= 10) iwinfmp->lpVtbl->setadpcmppz8emulate(iwinfmp, pctx->options.fmp.adpcmppz8emulate);
					iwinfmp->lpVtbl->Release(iwinfmp);
				}
			}
			break;
		}

		pctx->pifmpmd->lpVtbl->music_start(pctx->pifmpmd);
		return pctx;
	}
	if (pctx) Close(pctx);
	return 0;
}

unsigned Write(KMP_CTX *pctx, void *buf, unsigned smp)
{
	if (pctx->pifmpmd->lpVtbl->getloopcount(pctx->pifmpmd) == -1) return 0;
	pctx->pifmpmd->lpVtbl->getpcmdata(pctx->pifmpmd, buf, smp >> 2);
	return smp;
}

unsigned WriteSkip(KMP_CTX *pctx, unsigned smp)
{
	unsigned r = smp >> 2;
	short tmpbuf[512 * 2];
	if (pctx->pifmpmd->lpVtbl->getloopcount(pctx->pifmpmd) == -1) return 0;
	while (r > 512)
	{
		pctx->pifmpmd->lpVtbl->getpcmdata(pctx->pifmpmd, tmpbuf, 512);
		r -= 512;
	}
	if (r) pctx->pifmpmd->lpVtbl->getpcmdata(pctx->pifmpmd, tmpbuf, r);
	return smp;
}

unsigned SetPosition(KMP_CTX *pctx, unsigned ms)
{
	pctx->pifmpmd->lpVtbl->setpos(pctx->pifmpmd, ms);
	return pctx->pifmpmd->lpVtbl->getpos(pctx->pifmpmd);
}

int GetLoopFlag(KMP_CTX *pctx)
{
	return pctx->looplength != 0;
}

int GetSeekableFlag(KMP_CTX *pctx)
{
	return 1;
}

int GetMultisongFlag(KMP_CTX *pctx)
{
	return 0;
}

unsigned GetFrequency(KMP_CTX *pctx)
{
	return pctx->srate;
}
unsigned GetChannels(KMP_CTX *pctx)
{
	return 2;
}
unsigned GetSamplebits(KMP_CTX *pctx)
{
	return 16;
}
unsigned GetLength(KMP_CTX *pctx)
{
	if (!GetLoopFlag(pctx)) return pctx->length;
	return pctx->length + pctx->looplength * (pctx->loopcount - 1);
}
unsigned GetUnitSamples(KMP_CTX *pctx)
{
	return 44100 * 60 / 1000;	/* 60msec */
}

void XCRTINIT(void)
{
	unsigned pos;
	static const char *iniext = ".ini";
	pos = XDELEXT(iniFilePath, XGETMODULEPATH(), sizeof(iniFilePath));
	XSTRCPYN(iniFilePath + pos, iniext, sizeof(iniFilePath) - pos);
	rwork.mode = GetPrivateProfileBoolEx(sectionname, valuename_erm, TRUE, iniFilePath);
	rwork.srate_default = GetPrivateProfileIntEx(sectionname, valuename_sr, 44100, iniFilePath);
}

void XCRTTERM(void)
{
	DeinitSub(0);
	DeinitSub(1);
}

const char **GetPluginInfo(void)
{
	static const char *exts[3 + 5 + 3 + 1];
	static const char *description = "fmpmd.kpi v1.0" KPI_VERSIONS " by Mamiya";
	static char copyright[128] = "";
	static char flags[4] = "\0\0\0\0";
	const char **extp;
	int fmpversion, pmdversion;
	flags[0] = KPI_VERSION;
	flags[1] = rwork.mode;
	exts[0] = flags;
	exts[1] = description;
	exts[2] = copyright;
	pmdversion = GetDllVersion(0);
	fmpversion = GetDllVersion(1);
	if (!pmdversion && !fmpversion) return 0;
#if 1
	if (pmdversion && fmpversion)
		wsprintf(copyright, "PMDWin Version %d.%02d / WinFMP Version %d.%02d Copyright (C) C60", pmdversion / 100, pmdversion % 100, fmpversion / 100, fmpversion % 100);
	else if (pmdversion)
		wsprintf(copyright, "PMDWin Version %d.%02d / WinFMP is not installed. Copyright (C) C60", pmdversion / 100, pmdversion % 100);
	else
		wsprintf(copyright, "PMDWin is not installed. / WinFMP Version %d.%02d Copyright (C) C60", fmpversion / 100, fmpversion % 100);
#else
	{
		char *tmpp = copyright;
		XSTRCPYN(tmpp, "PMDWin ", MAX_PATH);
		tmpp += XSTRLEN(tmpp);
		if (pmdversion)
			wsprintf(tmpp, "Version %d.%02d", pmdversion / 100, pmdversion % 100);
		else
			wsprintf(tmpp, "is not installed.");
		tmpp += XSTRLEN(tmpp);
		XSTRCPYN(tmpp, " / Win FMP", MAX_PATH);
		tmpp += XSTRLEN(tmpp);
		if (fmpversion)
			wsprintf(tmpp, "Version %d.%02d", fmpversion / 100, fmpversion % 100);
		else
			wsprintf(tmpp, "is not installed.");
		tmpp += XSTRLEN(tmpp);
		XSTRCPYN(tmpp, " Copyright (C) C60", MAX_PATH);
	}
#endif
	extp = exts + 3;
	if (pmdversion)
	{
		*extp++ = ".M";
		*extp++ = ".M2";
		*extp++ = ".MZ";
		*extp++ = ".MP";
		*extp++ = ".MS";
	}
	if (fmpversion)
	{
		*extp++ = ".OPI";
		*extp++ = ".OVI";
		*extp++ = ".OZI";
	}
	*extp = 0;
	return exts;
}
